local E = unpack(select(2, ...)) -- Engine, Locale
local CO, NP = E:LoadModules("Config", "Nameplates")
NP.Autoload = true;

----------------------------------------
local _
local format		= string.format
local match			= string.match
----------------------------------------

--[[-------------------------------------------------

	This module modifies Blizzard nameplates
	
	The actual styling is done in the StylePlate method.
	The way nameplates work is that once a plate was created,
	it actually keeps its modifications.
	A modification of the styling preferences is possible at any point.

-------------------------------------------------]]--
-- Nameplate keys: Plate.UnitFrame, Plate.UnitFrame.healthBar, Plate.UnitFrame.healthBar.border,
--				   Plate.UnitFrame.castBar, Plate.UnitFrame.BuffFrame, Plate.UnitFrame.selectionHighlight
--				   Plate.UnitFrame.aggroHighlight, Plate.UnitFrame.LoseAggroAnim, Plate.UnitFrame.name
--				   Plate.UnitFrame.ClassificationFrame

-- Nameplate API: C_NamePlate.GetNamePlateForUnit, C_NamePlate.GetNamePlates, C_NamePlate.SetNamePlateEnemyClickThrough,
--				  C_NamePlate.SetNamePlateEnemySize, C_NamePlate.SetNamePlateFriendlyClickThrough, 
--				  C_NamePlate.SetNamePlateFriendlySize, C_NamePlate.SetNamePlateSelfClickThrough, 
--				  C_NamePlate.SetNamePlateSelfSize
local C_NamePlate_GetNamePlateForUnit 		= C_NamePlate.GetNamePlateForUnit
local C_NamePlate_SetNamePlateSelfSize 		= C_NamePlate.SetNamePlateSelfSize
local C_NamePlate_SetNamePlateEnemySize 	= C_NamePlate.SetNamePlateEnemySize
local C_NamePlate_SetNamePlateFriendlySize 	= C_NamePlate.SetNamePlateFriendlySize
local C_QuestLog_GetNumQuestLogEntries		= C_QuestLog.GetNumQuestLogEntries and C_QuestLog.GetNumQuestLogEntries or GetNumQuestLogEntries
--local C_QuestLog_GetTitleForLogIndex		= C_QuestLog.GetTitleForLogIndex and C_QuestLog.GetTitleForLogIndex or GetQuestLogTitle
local C_QuestLog_GetInfo 					= C_QuestLog.GetInfo
local C_QuestLog_IsComplete 				= C_QuestLog.IsComplete
local UnitIsTapDenied 	= UnitIsTapDenied
local UNKNOWN			= UNKNOWN

local NameFont_Exclusions = {["fontColor"] = true}
local LevelFont_Exclusions = {["fontColor"] = true}

local TappedColor = {0.5, 0.5, 0.5}
local ScanningTooltip = CreateFrame("GameTooltip", "CUI_NameplateScanningTooltip", nil, "GameTooltipTemplate")
local QuestListener = CreateFrame("Frame")
local Quests_Update

-- Kickable cast, Dispellable Auras, Configurable HP Text, Power Bar (If possible), Time to cast finish, Unit Name, Unit Level Text

-- Stores available nameplate unitIDs
NP.Plates = {}

--------------------------------------------------------
--	Functionality to actually access nameplates
--------------------------------------------------------
	
	-- Only gets called by the profile handler and config dialog
	function NP:LoadConfig()
		
		local Config = self.db
		local Unitframe
		
		for Name, _ in pairs(self.Plates) do
			
			Unitframe = self:GetPlateUnitframe(Name)
			
			if Unitframe and Unitframe.Mod then
				Unitframe.HPBar:SetSize(Config.barWidth, Config.barHeight)
				Unitframe.HPBar:SetStatusBarTexture(E.Media:Fetch("statusbar", Config.barTexture))
				
				if Config.questIcon.enable then
					if not Unitframe.QuestIcon then
						NP:InitQuestIcon(Unitframe)
					end
					
					if Unitframe.QuestIcon then
						Unitframe.QuestIcon:SetScale(Config.questIcon.scale)
						
						Unitframe.QuestIcon:ClearAllPoints()
						Unitframe.QuestIcon:SetPoint(E:InversePosition(Config.questIcon.position), Unitframe.HPBar, Config.questIcon.position, NP.db.questIcon.xOffset, Config.questIcon.yOffset)
						
						Unitframe.QuestIcon:ForceUpdate()
					end
				else
					if Unitframe.QuestIcon then
						Unitframe.QuestIcon:Hide()
					end
				end
				
				Unitframe:ForceUpdate()
			end
		end
		
		Quests_Update()
		
		C_NamePlate_SetNamePlateSelfSize(Config.clickableWidth, Config.clickableHeight)
		C_NamePlate_SetNamePlateEnemySize(Config.clickableWidth, Config.clickableHeight)
		C_NamePlate_SetNamePlateFriendlySize(Config.clickableWidth, Config.clickableHeight)
		
		-- CVar stuff
		E:RegisterCVar("nameplateShowSelf", nil, nil, true)
		E:RegisterCVar("nameplateShowAll", nil, nil, true)
	end
	
	function NP:GetPlateUnitframe(Unit)
		self.CurrentPlate = C_NamePlate_GetNamePlateForUnit(Unit)
		if self.CurrentPlate then
			return self.CurrentPlate.UnitFrame
		end
		
		return false
	end

	function NP:AddPlate(Unit)
		-- Do not alter player plate(s)
		if UnitIsUnit("player", Unit) then return end
		
		local Unitframe = self:GetPlateUnitframe(Unit)
		
		self.Plates[Unit] = true
		
		self:UpdatePlate(Unitframe)
	end

	function NP:RemovePlate(Unit)
		self.Plates[Unit] = nil
	end

	function NP:OnEvent(event, ...)	
		-- Add to table
		if event == "NAME_PLATE_UNIT_ADDED" then
			self:AddPlate(...)
		-- Remove from table
		elseif event == "NAME_PLATE_UNIT_REMOVED" then
			--self:RemovePlate(...)
		elseif event == "PLAYER_TARGET_CHANGED" then
			self:UpdateTargeted()
		
		-- This should fix Names not being updated properly
		elseif event == "UNIT_FACTION" or event == "UNIT_NAME_UPDATE" then
			self:UpdateGeneral(...)
		end
	end

--------------------------------------------------------
--	Nameplate styling
--------------------------------------------------------
	local OnlyShowQuests
	Quests_Update = function(event, questID)
		-- NP.QuestList = {
		-- 		[QuestName] = {[QuestID], [IsComplete]}
		-- }
		-- Start over fresh, just to be sure
		if not event or (event and event ~= 'QUEST_LOG_UPDATE') then
			NP.QuestList = NP.QuestList and wipe(NP.QuestList) or {}
		end
		
		local Title, IsHeader, IsCollapsedHeader, IsComplete, QuestID, IsTask
		OnlyShowQuests = NP.db.questIcon.onlyShowQuests
		local Info
		
		for i=1, C_QuestLog_GetNumQuestLogEntries() do
			Info = C_QuestLog_GetInfo(i)
			
			if Info then
				Title, IsHeader, IsCollapsedHeader, QuestID, IsTask = Info.title, Info.isHeader, Info.isCollapsed, Info.questID, Info.isTask
				
				if not IsHeader and (not (OnlyShowQuests and IsTask) or not OnlyShowQuests) then
					if QuestID then
						IsComplete = C_QuestLog_IsComplete(i)
					end
					NP.QuestList[Title] = NP.QuestList[Title] or {}
					
					NP.QuestList[Title].QuestID 	= QuestID
					NP.QuestList[Title].IsComplete 	= IsComplete
				end
			end
		end
		
		NP:UpdateAllQuestIcons()
	end
	
	local function Quests_OnEvent(self, event, ...)
		-- HANDLE QUEST EVENTS HERE
		Quests_Update(event, ...)
	end
	
	do
		E:RegisterEvents(QuestListener, 'QUEST_ACCEPTED', 'QUEST_REMOVED', 'QUEST_LOG_UPDATE', 'QUEST_FINISHED', 'PLAYER_ENTERING_WORLD')
		QuestListener:SetScript('OnEvent', Quests_OnEvent)
	end
	
	local function GetUnitColor(Unit)
		if not UnitIsTapDenied(Unit) then
			return E:GetUnitReactionColor(Unit, false)
		else
			return TappedColor
		end
	end
	
	local function Update_NameColor(self, newColor)
		if self.Mod and self.Name then
			
			newColor = newColor or GetUnitColor(self.Unit)
		
			self.name:Hide()
			self.Name:SetTextColor(unpack(newColor))
		end
	end
	
	local function Update_PlateColor(self)
		if not self.Mod then return end
		
		local Color = self.OverrideColor or GetUnitColor(self.Unit)
		
		self.HPBar:SetStatusBarColor(unpack(Color))
		Update_NameColor(self, Color)
	end
	
	local function Update_Health(self, event, ...)
		if not self.owner.Unit then return end
		local Max, Value = UnitHealthMax(self.owner.Unit), UnitHealth(self.owner.Unit)
		
		self:SetMinMaxValues(0, Max)
		self:SetValue(Value)
		if Max == 0 and Value == 0 then
			self.Font:SetText("Error")
		else
			self.Font:SetText(format("%.1f%%", (Value / Max) * 100))
		end
	end
	
	local function NameUpdate_Cancel(self)
		self:SetScript('OnUpdate', nil)
	end
	local function NameUpdate_OnUpdate(self, elapsed)
		-- Tick every 0.2 seconds
		self.Elapsed = (self.Elapsed or 0) + elapsed
		if self.Elapsed < 0.2 then return end
		self.Elapsed = 0
		
		-- Cancel scheduler when unit does not exist
		if not UnitExists(self.Owner.Unit) then
			NameUpdate_Cancel(self)
			return
		end
		
		self.NameCache = UnitName(self.Owner.Unit)
		
		-- If the name was found, update Text and Color
		if self.NameCache and self.NameCache ~= UNKNOWN then
			self.Name:SetText(self.NameCache)
			Update_PlateColor(self.Owner)
			
			-- Exit Scheduler
			NameUpdate_Cancel(self)
			return
		end
	end
	local function Schedule_NameUpdate(self)
		if not self:GetScript('OnUpdate') then			
			self:SetScript('OnUpdate', NameUpdate_OnUpdate)
		end
	end
	
	local function Update_NameFont(self)
		if self.Mod and self.Name then
			local Name = UnitName(self.Unit)
			self.Name:SetText(Name)
			
			if Name == UNKNOWN then
				Schedule_NameUpdate(self.FontContainer)
			end
		end
	end
	
	local function Update_LevelFont(self)
		local Level = UnitLevel(self.Unit)
		
		if self.Level.ShowAtMax == true and Level == E.UNIT_MAXLEVEL then
			self.Level:SetText(E.STR.EMPTY)
		else
			if Level ~= -1 then
				self.Level:SetText(Level)
				
				self.Level:SetTextColor(E:GetRGB(GetQuestDifficultyColor(Level)))
			else
				self.Level:SetText(E.STR.Boss)
				self.Level:SetTextColor(1, 0.2, 0.2)
			end
		end
	end
	
	local QuestIcon_ForceUpdate
	local function Construct_QuestIcon(self)
		self.QuestIcon = CreateFrame('Frame', 'CUI_NameplateQuestIcon', self.HPBar)
		self.QuestIcon:SetSize(12, 18)
		self.QuestIcon:SetScale(NP.db.questIcon.scale)
		self.QuestIcon:SetPoint(E:InversePosition(NP.db.questIcon.position), self.HPBar, NP.db.questIcon.position, NP.db.questIcon.xOffset, NP.db.questIcon.yOffset)
		self.QuestIcon.Tex = self.QuestIcon:CreateTexture(nil, 'OVERLAY')
		self.QuestIcon.Tex:SetAllPoints(self.QuestIcon)
		self.QuestIcon.Tex:SetTexture([[Interface\QUESTFRAME\AutoQuest-Parts]])
		self.QuestIcon.Tex:SetTexCoord(0.13476563 ,0.171875 ,0.015625 ,0.53125)
		
		self.QuestIcon.Owner = self
		self.QuestIcon.ForceUpdate = QuestIcon_ForceUpdate
	end
	
	local function Update_QuestMob(self)
		if not NP.db.questIcon.enable then
			if self.QuestIcon then
				self.QuestIcon:Hide()
			end
			return
		end
		
		ScanningTooltip:SetOwner(UIParent, 'ANCHOR_NONE')
		ScanningTooltip:SetUnit(self.Unit)
		ScanningTooltip:Show()
		
		self.QuestInfo = self.QuestInfo and wipe(self.QuestInfo) or {}
		
		local CurStr, IsMyQuest, Progress, QuestInfo
		for i=2,ScanningTooltip:NumLines() do			
			CurStr = _G['CUI_NameplateScanningTooltipTextLeft' .. i]
			CurStr = CurStr and CurStr:GetText()
			
			if not CurStr then break end
			if not QuestInfo then
				QuestInfo = NP.QuestList[CurStr]
			end
			
			-- /dump CUI[1]:LoadModule('Nameplates').QuestList
			
			if UnitIsPlayer(CurStr) and (IsInGroup() or IsInRaid()) then
				IsMyQuest = CurStr == E.PlayerName
			elseif QuestInfo then
				self.QuestInfo.ObjectiveCount 				= self.QuestInfo.ObjectiveCount or {}
				
				local Current, Goal = match(CurStr, '(%d+)/(%d+)')
				if Current and Goal then
					self.QuestInfo.ObjectiveCount.Current 	= Current
					self.QuestInfo.ObjectiveCount.Goal 		= Goal
					self.QuestInfo.ObjectiveCount.Finished 	= Current == Goal
				else
					self.QuestInfo.ObjectiveCount.Finished	= QuestInfo.IsComplete
				end
				
				self.QuestInfo.IsComplete 					= QuestInfo.IsComplete
				self.QuestInfo.QuestID 						= QuestInfo.QuestID
			end
		end
		
		ScanningTooltip:Hide()
		
		if self.QuestInfo and self.QuestInfo.QuestID and not self.QuestInfo.IsComplete and not self.QuestInfo.ObjectiveCount.Finished then
			if not self.QuestIcon then
				Construct_QuestIcon(self)
			end
			
			self.QuestIcon:Show()
		else
			if self.QuestIcon then
				self.QuestIcon:Hide()
			end
		end
	end
	
	local function Nameplate_OnShow(self)
		NP:UpdatePlate(self)
	end
	
	local function ForceUpdate(self)
		NP:UpdatePlate(self)
	end
	
	QuestIcon_ForceUpdate = function(self)
		Update_QuestMob(self.Owner)
	end
	
	function NP:InitQuestIcon(Unitframe)
		Update_QuestMob(Unitframe)
	end
	
	function NP:UpdatePlate(Unitframe)
		
		Unitframe.Unit = Unitframe.unit
		
		if not Unitframe.Mod then
			self:StylePlate(Unitframe, Unitframe.Unit)
			Unitframe.ForceUpdate = ForceUpdate
		end
		
		Update_Health(Unitframe.HPBar)
		
		Update_NameFont(Unitframe)
		Update_LevelFont(Unitframe)
		
		Update_QuestMob(Unitframe)
		
		-- Update color last, so we can make changes to it
		Update_PlateColor(Unitframe)
		self:UpdateTargeted()
	end
	
	function NP:UpdateAllQuestIcons()
		local Unitframe
		for Name, _ in pairs(self.Plates) do
			Unitframe = self:GetPlateUnitframe(Name)
			
			if Unitframe and Unitframe.QuestIcon and Unitframe:IsVisible() then
				Unitframe.QuestIcon:ForceUpdate()
			end
		end
	end
	
	function NP:StylePlate(Unitframe, Unit)
		-- Only mod each Plate once
		if not Unit then return end
		
		if not Unitframe.Mod then
			---------------------------
			-- Healthbar
				-- Bar smoothing
					
					-- This actually is a completely new bar so we never have to mess with the Blizz ones [Taint 'n stuff]
					
					local Healthbar = CreateFrame("statusbar", "NamePlate_HealthBar" .. Unit, Unitframe)
					Healthbar:SetScript("OnEvent", Update_Health)
					Healthbar:RegisterEvent("UNIT_HEALTH")
					if not E:IsRunningShadowlands() then
						Healthbar:RegisterEvent("UNIT_HEALTH_FREQUENT")
					end
					
					Unitframe.HPBar = Healthbar
					
					Healthbar.owner = Unitframe
					
					Healthbar.Background = E:CreateBackground(Healthbar)
					Healthbar.Border = E:CreateBorder(Healthbar)
					
					E.Libs.LibSmooth:SmoothBar(Healthbar)
					
				-- Bar texture
					E:RegisterStatusBar(Healthbar)
					Healthbar:SetStatusBarTexture(E.Media:Fetch("statusbar", self.db.barTexture))
					
					Healthbar:SetSize(self.db.barWidth, self.db.barHeight)
					Healthbar:SetPoint("CENTER", Unitframe, "CENTER", 0, 0)
					
					-- Keep old bar hidden
					Unitframe.healthBar:HookScript('OnShow', function(self) self:Hide() end)
					Unitframe.healthBar:Hide()
			
			
			---------------------------
			-- @TODO: We actually can register those fonts via PathFont.
			-- 		  Also, pack the fonts into a compact and simple table
			-- Fonts
			
				local FontContainer = CreateFrame("Frame", nil, Healthbar)
				FontContainer:SetAllPoints(true)
				FontContainer:SetScript("OnEvent", Fonts_OnEvent)
				FontContainer.Owner = Unitframe
				Unitframe.FontContainer = FontContainer
			
				-- Name
					
					local Name = FontContainer:CreateFontString(nil)
						E:InitializeFontFrame(Name, "OVERLAY", "FRIZQT__.TTF", 9, {0.9, 0.9, 0.9}, 0.9, {0, -4}, "", 0, 0, FontContainer, "CENTER", {1,1})
					FontContainer:RegisterEvent("UNIT_NAME_UPDATE")
					Unitframe.Name, FontContainer.Name = Name, Name
				
				-- Level
					
					local Level = FontContainer:CreateFontString(nil)
						E:InitializeFontFrame(Level, "OVERLAY", "FRIZQT__.TTF", 9, {0.9, 0.9, 0.9}, 0.9, {0, -4}, "", 0, 0, FontContainer, "LEFT", {1,1})
					FontContainer:RegisterEvent("UNIT_LEVEL")
					Unitframe.Level, FontContainer.Level = Level, Level
					
				-- Health
					
					local Health = FontContainer:CreateFontString(nil)
						E:InitializeFontFrame(Health, "OVERLAY", "FRIZQT__.TTF", 8, {0.9, 0.9, 0.3}, 0.9, {-3, 0}, "", 0, 0, FontContainer, "RIGHT", {1,1})						
					Unitframe.Health, FontContainer.Health = Health, Health
						
					E:RegisterPathFont(Health, "db.profile.nameplates.health")
					E:UpdatePathFont("db.profile.nameplates.health")
					
					E:RegisterPathFont(Level, "db.profile.nameplates.level", LevelFont_Exclusions)
					E:UpdatePathFont("db.profile.nameplates.level")	
					
					E:RegisterPathFont(Name, "db.profile.nameplates.name", NameFont_Exclusions)
					E:UpdatePathFont("db.profile.nameplates.name")
					
				
			---------------------------
			-- Various stuff to make our lives easier
			
				Unitframe.HealthFont	 	= Health
				Healthbar.Font 				= Health
				hooksecurefunc("CompactUnitFrame_UpdateHealthColor", Update_PlateColor)
				hooksecurefunc("CompactUnitFrame_UpdateName", Update_NameColor)
				
			-- Selection texture
				Unitframe.selectionHighlight:ClearAllPoints()
				Unitframe.selectionHighlight:SetTexture(nil)
				Unitframe.healthBar.border:Hide()
			
			-- Custom selection indicator
				Unitframe.SelectionBorder = E:CreateBorder(Healthbar, nil, 1)
				Unitframe.SelectionBorder:SetBackdropBorderColor(1, 1, 1, 1)
				Unitframe.SelectionBorder:Hide()
				
			-- Classification Icon
				Unitframe.ClassificationFrame:ClearAllPoints()
				Unitframe.ClassificationFrame:SetPoint("BOTTOMRIGHT", Healthbar, "TOPLEFT")
				
			-- Force update when it shows up
			-- This is to prevent a bug with spawned units whose name cannot be resolved
				Unitframe:SetScript("OnShow", Nameplate_OnShow)
			
			
			---------------------------
			Unitframe.Mod = true
		end
	end
	
	function NP:UpdateTargeted()
		for unit, _ in pairs(self.Plates) do
			
			self.TargetPlate = NP:GetPlateUnitframe(unit)
			
			if type(self.TargetPlate) == "table" then
				if self.TargetPlate.SelectionBorder then
					if UnitExists("target") and UnitIsUnit(self.TargetPlate.Unit, "target") then
						self.TargetPlate.SelectionBorder:Show()
					else
						self.TargetPlate.SelectionBorder:Hide()
					end
				end
			end
		end
	end
	
	-- This seems to get called whenever we have to perform a critical update (Name etc)
	function NP:UpdateGeneral(Unit)
		local Plate = self:GetPlateUnitframe(Unit)
		if Plate then
			NP:UpdatePlate(Plate)
		end
	end



--------------------------------
function NP:UpdateDB()
	self.db = CO.db.profile.nameplates
end
function NP:Init()
	
	self:UpdateDB()
	E:RegisterFontExclusions("db.profile.nameplates.level", LevelFont_Exclusions)
	E:RegisterFontExclusions("db.profile.nameplates.name", NameFont_Exclusions)
	
	if not CO.db.char.nameplates.enable then return end
	
	self:RegisterEvent("NAME_PLATE_UNIT_ADDED")
	self:RegisterEvent("NAME_PLATE_UNIT_REMOVED")
	self:RegisterEvent("PLAYER_TARGET_CHANGED")
	self:RegisterEvent("UNIT_FACTION")
	
	self:SetScript("OnEvent", NP.OnEvent)
	self:LoadConfig()
end
E:AddModule("Nameplates", NP)